# MIT License
#
# Copyright (c) 2023 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from datetime import datetime
from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional

from mkdocs.config.defaults import MkDocsConfig

# noinspection PyProtectedMember
from mkdocs_publisher._shared import mkdocs_utils
from mkdocs_publisher.blog.config import BlogPluginConfig
from mkdocs_publisher.meta.config import MetaPluginConfig


@dataclass
class BlogPost:
    """Single blog post data container"""

    title: str
    date: datetime
    path: Optional[str]
    content: Optional[str]
    tags: Optional[List[str]]
    categories: Optional[List[str]]
    slug: Optional[str] = None
    teaser: str = ""
    is_teaser: bool = False

    @property
    def as_dict(self) -> dict:
        return asdict(self)


@dataclass
class Translation:
    teaser_link_text: str
    blog_page_title: str
    blog_navigation_name: str
    recent_blog_posts_navigation_name: str
    archive_page_title: str
    archive_navigation_name: str
    categories_page_title: str
    categories_navigation_name: str
    tags_page_title: str
    tags_navigation_name: str
    newer_posts: str
    older_posts: str

    @property
    def as_dict(self) -> dict:
        return asdict(self)


@dataclass
class BlogConfig:
    mkdocs_config: MkDocsConfig = field(init=False)
    plugin_config: BlogPluginConfig = field(init=False)
    meta_config: Optional[MetaPluginConfig] = field(init=False, default=None)
    translation: Translation = field(init=False)
    temp_dir: Path = field(init=False)
    docs_dir: Path = field(init=False)
    blog_dir: Path = field(init=False)
    site_dir: Path = field(init=False)
    blog_posts: Dict[datetime, BlogPost] = field(init=False, default_factory=lambda: dict())
    temp_files: Dict[str, Path] = field(init=False, default_factory=lambda: dict())

    @property
    def temp_files_list(self) -> List[str]:
        temp_files = []
        for path in self.temp_files.values():
            temp_files.append(str(path.relative_to(self.temp_dir)))
        return temp_files

    def parse_configs(self, mkdocs_config: MkDocsConfig, plugin_config: BlogPluginConfig):
        from mkdocs_publisher.blog.translate import Translate

        self.mkdocs_config = mkdocs_config
        self.plugin_config = plugin_config
        self.meta_config: Optional[MetaPluginConfig] = mkdocs_utils.get_plugin_config(
            mkdocs_config=mkdocs_config, plugin_name="pub-meta"
        )  # type: ignore
        self.temp_dir = Path(plugin_config.temp_dir)
        self.docs_dir = Path(mkdocs_config.docs_dir)
        self.blog_dir = Path(plugin_config.blog_dir)
        self.site_dir = Path(mkdocs_config.site_dir)
        self.translation = Translate(config=plugin_config).translation

        self.temp_dir.mkdir(exist_ok=True)



# MIT License
#
# Copyright (c) 2023 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from multiprocessing import cpu_count
from pathlib import Path
from typing import Dict
from typing import Literal
from typing import cast

import yaml
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin
from mkdocs.plugins import event_priority

from mkdocs_publisher.minifier import minifiers
from mkdocs_publisher.minifier.base import CachedFile
from mkdocs_publisher.minifier.config import MinifierConfig

log = logging.getLogger("mkdocs.plugins.publisher.minifier.plugin")


class MinifierPlugin(BasePlugin[MinifierConfig]):
    def __init__(self):
        self._on_serve: bool = False

    def on_startup(self, *, command: Literal["build", "gh-deploy", "serve"], dirty: bool) -> None:
        if command == "serve":
            self._on_serve = True

    @event_priority(-100)  # Run after all other plugins
    def on_post_build(self, *, config: MkDocsConfig) -> None:

        Path(self.config.cache_dir).mkdir(exist_ok=True)

        # TODO: Add path to tools checker
        cached_files: Dict[str, CachedFile] = {}

        if self.config.threads == 0:
            self.config.threads = int(cpu_count())
        if self.config.threads < 1:
            log.info('Number of "threads" cannot be smaller than 1 (changing to minimal 1)')
            self.config.threads = 1
        log.info(f"Threads used for minifiers: {self.config.threads}")

        cached_files_list: Path = Path(self.config.cache_dir) / self.config.cache_file
        if cached_files_list.exists():
            try:
                with open(cached_files_list, "r") as yaml_file:
                    cached_files = yaml.safe_load(yaml_file)
                    for file_path, cached_file in cached_files.items():
                        cached_files[file_path] = CachedFile(**cast(dict, cached_file))
            except (yaml.YAMLError, AttributeError) as e:
                log.warning(f"File '{cached_files_list}' corrupted. Rebuilding cache.")
                log.debug(e)

        if self.config.png.enabled and (
            (not self._on_serve) or (self._on_serve and self.config.png.enabled_on_serve)
        ):
            minifiers.PngMinifier(
                plugin_config=self.config, mkdocs_config=config, cached_files=cached_files
            )()

        if self.config.jpeg.enabled and (
            (not self._on_serve) or (self._on_serve and self.config.jpeg.enabled_on_serve)
        ):
            minifiers.JpegMinifier(
                plugin_config=self.config, mkdocs_config=config, cached_files=cached_files
            )()

        if self.config.svg.enabled and (
            (not self._on_serve) or (self._on_serve and self.config.svg.enabled_on_serve)
        ):
            minifiers.SvgMinifier(
                plugin_config=self.config, mkdocs_config=config, cached_files=cached_files
            )()

        if self.config.html.enabled and (
            (not self._on_serve) or (self._on_serve and self.config.html.enabled_on_serve)
        ):
            minifiers.HtmlMinifier(
                plugin_config=self.config, mkdocs_config=config, cached_files=cached_files
            )()

        if self.config.css.enabled and (
            (not self._on_serve) or (self._on_serve and self.config.css.enabled_on_serve)
        ):
            minifiers.CssMinifier(
                plugin_config=self.config, mkdocs_config=config, cached_files=cached_files
            )()

        if self.config.js.enabled and (
            (not self._on_serve) or (self._on_serve and self.config.js.enabled_on_serve)
        ):
            minifiers.JsMinifier(
                plugin_config=self.config, mkdocs_config=config, cached_files=cached_files
            )()

        cached_files_for_yaml = {}
        for file_name, cached_file in cached_files.items():
            # TODO: Cleanup yaml from unused/deleted files and delete those that are unused
            cached_file = {k: str(v) for k, v in cached_file.as_dict.items()}
            cached_files_for_yaml[file_name] = cached_file

        with open(cached_files_list, "w") as yaml_file:
            yaml.safe_dump(cached_files_for_yaml, yaml_file)



# MIT License
#
# Copyright (c) 2023 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from collections import OrderedDict
from datetime import datetime
from pathlib import Path
from typing import Dict
from typing import cast

import jinja2
import yaml
from mkdocs.structure.files import File
from mkdocs.structure.files import Files

# noinspection PyProtectedMember
from mkdocs_publisher._shared import resources

# noinspection PyProtectedMember
from mkdocs_publisher._shared.urls import slugify
from mkdocs_publisher.blog.structures import BlogConfig

log = logging.getLogger("mkdocs.plugins.publisher.blog.creators")


def create_blog_files(
    blog_config: BlogConfig,
    files: Files,
):
    for temp_file in blog_config.temp_files.values():
        try:
            file = File(
                path=str(Path(temp_file).relative_to(blog_config.temp_dir)),
                src_dir=str(blog_config.temp_dir),
                dest_dir=str(blog_config.site_dir),
                use_directory_urls=blog_config.mkdocs_config.use_directory_urls,
            )
            files.append(file)
        except ValueError:
            pass


def create_blog_post_pages(
    start_page: bool,
    blog_config: BlogConfig,
    config_nav: OrderedDict,
) -> None:
    """Create blog posts index files."""

    log.info("Creating blog posts index files")
    posts_chunks: Dict[str, list] = {}
    archive_chunks: Dict[str, list] = {}
    categories_chunks: Dict[str, list] = {}
    tags_chunks: Dict[str, list] = {}

    # Build post index pages
    for index, date in enumerate(sorted(blog_config.blog_posts, reverse=True)):
        index = (
            "index"
            if start_page and index < blog_config.plugin_config.posts_per_page
            else f"index-{str(index//blog_config.plugin_config.posts_per_page)}"
        )
        if index not in posts_chunks:
            posts_chunks[index] = []
        posts_chunks[index].append(blog_config.blog_posts[date])

    # Build archive, category and tag pages
    for date in sorted(blog_config.blog_posts, reverse=True):
        date: datetime
        year = str(date.year)
        if year not in archive_chunks:
            archive_chunks[year] = []
        archive_chunks[year].append(blog_config.blog_posts[date].as_dict)

        categories = cast(str, blog_config.blog_posts[date].categories)
        for category in categories:
            if category not in categories_chunks:
                categories_chunks[category] = []
            categories_chunks[category].append(blog_config.blog_posts[date].as_dict)

        tags = cast(list, blog_config.blog_posts[date].tags)
        for tag in tags:
            if tag not in tags_chunks:
                tag = tag.replace("/", "-")
                tags_chunks[tag] = []
            tags_chunks[tag].append(blog_config.blog_posts[date].as_dict)

    # Reorder categories alphabetically
    categories_chunks = {cat: categories_chunks[cat] for cat in sorted(categories_chunks)}

    # Reorder tags alphabetically
    tags_chunks = {tag: tags_chunks[tag] for tag in sorted(tags_chunks)}
    config_nav[blog_config.translation.blog_navigation_name] = []

    for key, single_posts_chunk in posts_chunks.items():
        file_name = f"{key}.md"
        file_path = blog_config.temp_dir / file_name

        _render_and_write_page(
            single_posts_chunk=single_posts_chunk,
            file_path=file_path,
            blog_config=blog_config,
            page_title=blog_config.translation.blog_page_title,
        )

        blog_config.temp_files[f"{blog_config.translation.blog_navigation_name}/{key}"] = file_path
        log.debug(f"Creating blog post chunk file: {file_path}")

        config_nav[blog_config.translation.blog_navigation_name].append({key: file_name})

    config_nav[blog_config.translation.blog_navigation_name].append(
        {
            blog_config.translation.archive_navigation_name: _create_pages(
                blog_config=blog_config,
                posts_chunks=archive_chunks,
                sub_dir=Path(blog_config.plugin_config.archive_subdir),
                page_title=blog_config.translation.archive_page_title,
            )
        }
    )

    config_nav[blog_config.translation.blog_navigation_name].append(
        {
            blog_config.translation.categories_navigation_name: _create_pages(
                blog_config=blog_config,
                posts_chunks=categories_chunks,
                sub_dir=Path(blog_config.plugin_config.categories_subdir),
                page_title=blog_config.translation.categories_page_title,
            )
        }
    )

    config_nav[blog_config.translation.blog_navigation_name].append(
        {
            blog_config.translation.tags_navigation_name: _create_pages(
                blog_config=blog_config,
                posts_chunks=tags_chunks,
                sub_dir=Path(blog_config.plugin_config.tags_subdir),
                page_title=blog_config.translation.tags_page_title,
            )
        }
    )


def _create_pages(
    blog_config: BlogConfig,
    posts_chunks: Dict[str, list],
    sub_dir: Path,
    page_title: str,
) -> list[dict[str, str]]:
    config_nav = []

    blog_temp_dir = blog_config.temp_dir / blog_config.plugin_config.blog_dir
    blog_temp_dir.mkdir(exist_ok=True)

    pages_dir = blog_temp_dir / sub_dir
    pages_dir.mkdir(exist_ok=True)

    for key, single_posts_chunk in posts_chunks.items():
        file_name = f"{key}.md"
        file_path = pages_dir / file_name
        _render_and_write_page(
            single_posts_chunk=single_posts_chunk,
            file_path=file_path,
            blog_config=blog_config,
            page_title=f"{page_title} - {key}",
        )

        blog_config.temp_files[f"{sub_dir}/{key}"] = file_path
        log.debug(f"Creating blog post chunk file: {file_path}")

        config_nav.append({key: f"{blog_config.plugin_config.blog_dir}/{sub_dir}/{file_name}"})
    return config_nav


def _render_and_write_page(
    single_posts_chunk: list,
    file_path: Path,
    blog_config: BlogConfig,
    page_title: str,
):
    # TODO: Add templates from override
    # templates = jinja2.Environment(loader=jinja2.FileSystemLoader("templates/"))
    # print(templates.list_templates())
    # template = templates.get_template("index.html")

    index_template = resources.read_template_file(template_file_name="posts-list.html")
    context = {
        "posts": single_posts_chunk,
        "site_url": str(blog_config.mkdocs_config.site_url),
        "config": blog_config.plugin_config,
        "translation": blog_config.translation,
    }
    template = jinja2.Environment(loader=jinja2.BaseLoader()).from_string(index_template)
    markdown = template.render(context)

    # TODO: when using pub-meta key name should be taken from plugin config
    page_meta = {"title": page_title}
    # TODO: consider moving slugify configuration to mkdocs.yaml
    if file_path.name.startswith("index-0"):
        slug = blog_config.plugin_config.slug
    elif file_path.name.startswith("index"):
        slug = f"{blog_config.plugin_config.slug}/{file_path.stem.split('-')[-1]}"
    else:
        slug = slugify(text=page_title.split("-")[-1].strip())
    page_meta["slug"] = slug
    page_meta["publish"] = "published"
    if not blog_config.plugin_config.searchable_non_posts:
        page_meta["search"] = {"exclude": True}  # type: ignore

    with open(file_path, mode="w") as teasers_index:
        teasers_index.write(f"---\n{yaml.dump(page_meta)}\n---\n\n{markdown}")



# MIT License
#
# Copyright (c) 2023 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from datetime import datetime
from pathlib import Path
from typing import Literal
from typing import Optional
from typing import cast
from urllib.parse import quote

from mkdocs.config import Config
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin
from mkdocs.plugins import event_priority
from mkdocs.structure.files import Files
from mkdocs.structure.nav import Link
from mkdocs.structure.nav import Navigation
from mkdocs.structure.nav import Section
from mkdocs.structure.pages import Page
from mkdocs.utils import meta as meta_parser

# noinspection PyProtectedMember
from mkdocs_publisher._shared import mkdocs_utils
from mkdocs_publisher.blog.config import BlogPluginConfig
from mkdocs_publisher.meta.config import MetaPluginConfig
from mkdocs_publisher.meta.config import _MetaPublishConfig
from mkdocs_publisher.obsidian.config import ObsidianPluginConfig

log = logging.getLogger("mkdocs.plugins.publisher.meta.plugin")
INDEX_FILE_NAME = "index.md"


class MetaPlugin(BasePlugin[MetaPluginConfig]):
    def __init__(self):
        """
        TODO: SEO optimizations
        Useful links for sitemap.xml manipulation etc.:
        - https://octamedia.pl/blog/mapa-strony-xml/
        - https://octamedia.pl/blog/linkowanie-wewnetrzne/ (useful for obsidian backlinks)
        """
        self._blog_config: Optional[BlogPluginConfig] = None
        self._dirs_slugs: dict = {}
        self._dirs_titles: dict = {}
        self._draft_dirs: list = []
        self._hidden_dirs: list = []
        self._draft_files: list = []
        self._hidden_files: list = []
        self._not_other_files: list = []
        self._on_serve = False

    def _build_config_nav(self, directory: Path, relative_to: Path) -> list:
        nav = list()
        sorted_files = sorted([f for f in directory.glob("*")])

        # Make index.md a first file in given directory
        other_files = [f for f in sorted_files if f.name not in self._not_other_files]
        sorted_files = [f for f in sorted_files if f.name == "index.md"]
        sorted_files.extend(other_files)

        for file in sorted_files:
            if not any([d for d in self._draft_dirs if file.is_relative_to(d)]):
                if file.is_dir():
                    title = self._dirs_titles.get(file, file.stem)
                    dir_nav = self._build_config_nav(directory=file, relative_to=relative_to)
                    if isinstance(dir_nav, list) and len(dir_nav) > 0:  # Skip empty directories
                        nav.append({title: dir_nav})
                elif file.is_file() and file.suffix == ".md":
                    with file.open(encoding="utf-8-sig", errors="strict") as md_file:
                        markdown, meta = meta_parser.get_data(md_file.read())
                        title = meta.get(self.config.title.key_name, file.stem)

                        # Read document publication status
                        publish = meta.get(self.config.publish.key_name)
                        if publish is None:
                            log_message = (
                                f'Missing "{self.config.publish.key_name}" value in '
                                f'file "{file.relative_to(relative_to)}". Setting to '
                                f'default value: "{self.config.publish.file_default}".'
                            )
                            if self.config.publish.file_warn_on_missing:
                                log.warning(log_message)
                            else:
                                log.debug(log_message)
                            publish = self.config.publish.file_default

                        # TODO: make it configurable
                        if publish in ["draft", "false", False] and self._on_serve:
                            publish = "published"

                        # Add file to variables used in nav cleanup based on publication status
                        if publish in ["draft", "false", False]:
                            self._draft_files.append(str(file.relative_to(relative_to)))
                        elif publish == "hidden":
                            self._hidden_files.append(str(file.relative_to(relative_to)))

                        # Add not draft file to navigation (hidden will be removed in nav cleanup)
                        if publish not in ["draft", "false", False]:
                            nav.append({title: str(file.relative_to(relative_to))})
            elif (
                file.is_dir()
                and self._blog_config is not None
                and str(file.relative_to(relative_to)) == self._blog_config.blog_dir
            ):
                nav.append({file.stem: str(file.relative_to(relative_to))})
        return nav

    def _nav_cleanup(self, items: list, elements_to_remove: list) -> list:

        nav = []
        for item in items:
            if isinstance(item, Page) and Path(item.file.abs_src_path) not in elements_to_remove:
                nav.append(item)
            elif isinstance(item, Link):
                nav.append(item)
            elif isinstance(item, Section):
                item.children = self._nav_cleanup(
                    items=item.children, elements_to_remove=elements_to_remove
                )
                if len(item.children) > 0:  # Skip empty Sections
                    nav.append(item)

        return nav

    def on_startup(self, *, command: Literal["build", "gh-deploy", "serve"], dirty: bool) -> None:
        if command == "serve":
            self._on_serve = True

    @event_priority(100)  # Run before any other plugins
    def on_config(self, config: MkDocsConfig) -> Optional[Config]:

        # Setup some default values
        self._not_other_files = [INDEX_FILE_NAME, self.config.dir_meta_file]

        log.info(f'Reading meta data from "{self.config.dir_meta_file}" files')
        for meta_file in Path(config.docs_dir).glob(f"**/{self.config.dir_meta_file}"):
            with meta_file.open(encoding="utf-8-sig", errors="strict") as md_file:

                # Add empty line at the end of file and read all data
                markdown, meta = meta_parser.get_data(f"{md_file.read()}\n")

                # Read slug for directories
                slug = meta.get(self.config.slug.key_name)
                if slug is not None:
                    self._dirs_slugs[meta_file.parent] = slug

                # Read title for directory
                title = meta.get(self.config.title.key_name)
                if title is not None:
                    self._dirs_titles[meta_file.parent] = title

                # Read directories publication status
                publish = meta.get(str(self.config.publish.key_name))
                if publish is None:
                    if self.config.publish.dir_warn_on_missing:
                        log.warning(
                            f'Missing "{self.config.publish.key_name}" value in '
                            f'file "{meta_file.relative_to(config.docs_dir)}". Setting to '
                            f'default value: "{self.config.publish.dir_default}".'
                        )
                    publish = self.config.publish.dir_default
                dir_path = str(meta_file.parent.relative_to(config.docs_dir))

                if publish in ["draft", "false", False] and self._on_serve:
                    publish = "published"

                if publish in ["draft", "false", False]:
                    self._draft_dirs.append(dir_path)
                elif publish == "hidden":
                    self._hidden_dirs.append(dir_path)
                elif publish not in _MetaPublishConfig.dir_default.choices:  # type: ignore
                    log.warning(
                        f'Wrong key "{self.config.publish.key_name}" value '
                        f'in file "{meta_file.relative_to(config.docs_dir)}" (only '
                        f"{_MetaPublishConfig.dir_default.choices} are possible)"  # type: ignore
                    )

                log.debug(
                    f"Title: {title}, publish: {publish}, "
                    f"slug: {slug} "
                    f'(directory: "{dir_path}")'
                )

        # Add blog dir to one that will be skipped (blog has its own file resolution order)
        self._blog_config: Optional[BlogPluginConfig] = mkdocs_utils.get_plugin_config(
            mkdocs_config=config,
            plugin_name="pub-blog",
        )  # type: ignore
        if self._blog_config is not None:
            blog_dir = cast(BlogPluginConfig, self._blog_config).blog_dir
            # TODO: cleanup mess with short directory names in self._draft_dirs
            if (
                blog_dir not in self._draft_dirs
                and Path(config.docs_dir) / blog_dir not in self._draft_dirs
            ):  # type: ignore
                self._draft_dirs.append(blog_dir)

        # Add obsidian dirs to ones that will be skipped
        obsidian_config: Optional[ObsidianPluginConfig] = mkdocs_utils.get_plugin_config(
            mkdocs_config=config,
            plugin_name="pub-obsidian",
        )  # type: ignore
        if obsidian_config is not None:
            # TODO: cleanup mess with short directory names in self._draft_dirs
            if (
                obsidian_config.obsidian_dir != ""
                and obsidian_config.obsidian_dir not in self._draft_dirs
                and Path(config.docs_dir) / obsidian_config.obsidian_dir not in self._draft_dirs
            ):
                self._draft_dirs.append(obsidian_config.obsidian_dir)
            if (
                obsidian_config.templates_dir != ""
                and obsidian_config.templates_dir not in self._draft_dirs
                and Path(config.docs_dir) / obsidian_config.templates_dir not in self._draft_dirs
            ):
                self._draft_dirs.append(obsidian_config.templates_dir)
        self._draft_dirs = [Path(config.docs_dir) / d for d in self._draft_dirs]
        self._hidden_dirs = [Path(config.docs_dir) / f for f in self._hidden_dirs]
        for draft_dir in self._draft_dirs:
            if not draft_dir.exists():
                log.warning(
                    f'Directory "{draft_dir.relative_to(config.docs_dir)}" doesn\'t exists'
                )
        config.nav = self._build_config_nav(
            directory=Path(config.docs_dir), relative_to=Path(config.docs_dir)
        )
        log.info(
            f"Draft directories: "
            f"{[str(d.relative_to(config.docs_dir)) for d in self._draft_dirs]}"
        )
        log.info(
            f"Hidden directories: "
            f"{[str(d.relative_to(config.docs_dir)) for d in self._hidden_dirs]}"
        )
        draft_files_list = [
            str(f.relative_to(config.docs_dir)) if isinstance(f, Path) else f
            for f in self._draft_files
        ]
        log.info(f"Draft files: {draft_files_list}")
        hidden_files_list = [
            str(f.relative_to(config.docs_dir)) if isinstance(f, Path) else f
            for f in self._hidden_files
        ]
        log.info(f"Hidden files: {hidden_files_list}")

        return config

    @event_priority(-100)
    def on_files(self, files: Files, *, config: MkDocsConfig) -> Optional[Files]:

        relative_draft_dirs = [str(f.relative_to(config.docs_dir)) for f in self._draft_dirs]
        if self._blog_config is not None:
            relative_draft_dirs.remove(self._blog_config.blog_dir)

        new_files = Files(files=[])
        for file in files:

            file_path = Path(file.abs_src_path)
            if file_path.suffix == ".md":
                if (
                    any([d for d in self._draft_dirs if file_path.is_relative_to(d)])
                    and Path(file_path) not in self._draft_files
                ):
                    self._draft_files.append(file_path)
                elif (
                    any([d for d in self._hidden_dirs if file_path.is_relative_to(d)])
                    and Path(file_path) not in self._hidden_files
                ):
                    self._hidden_files.append(file_path)

            if (
                not any([file.src_uri.startswith(draft_dir) for draft_dir in relative_draft_dirs])
                and str(Path(file.src_uri).name) != self.config.dir_meta_file
                and str(file.src_uri) not in self._draft_files
            ):
                new_files.append(file)

                if file.is_documentation_page():

                    meta_file = Path(file.abs_src_path)
                    with meta_file.open(encoding="utf-8-sig", errors="strict") as md_file:
                        markdown, meta = meta_parser.get_data(md_file.read())

                        # Read slug
                        slug = meta.get(self.config.slug.key_name)
                        # TODO: add slugify and config
                        if slug is None and self.config.slug.warn_on_missing:
                            log.warning(
                                f'File "{file.src_path}" has no '
                                f'"{self.config.slug.key_name}" meta data'
                            )

                        # Get URL parts
                        if file.url.endswith("/"):
                            file.url = file.url[0:-1]
                        url_parts = file.url.split("/")

                        # Get abs file parts
                        path_parts: list[Path] = []
                        for path_part in file.src_path.split("/"):
                            if not path_parts:
                                path_parts.append(Path(config.docs_dir) / path_part)
                            else:
                                path_parts.append(path_parts[-1] / path_part)

                        # Replace URL parts that have slug defined
                        for position, path_part in enumerate(path_parts):
                            if path_part in self._dirs_slugs:
                                url_parts[position] = self._dirs_slugs[path_part]

                        # Replace last URL part with current file slug if exists
                        if slug is not None and self.config.slug.enabled:
                            url_parts[-1] = slug

                        # Recreate file params based on URL with replaced parts
                        if file.url != ".":  # Do not modify main index page
                            file.url = quote(f"{'/'.join(url_parts)}/")
                            url_parts.append(file.dest_uri.split("/")[-1])
                            file.dest_uri = quote("/".join(url_parts))
                            file.abs_dest_path = str(Path(config.site_dir) / file.dest_uri)

        return new_files

    def on_nav(
        self, nav: Navigation, *, config: MkDocsConfig, files: Files
    ) -> Optional[Navigation]:

        elements_to_remove = []
        elements_to_remove.extend(self._draft_files)
        elements_to_remove.extend(self._hidden_files)

        nav.items = self._nav_cleanup(items=nav.items, elements_to_remove=elements_to_remove)

        return nav

    @event_priority(-100)  # Run after all other plugins
    def on_page_markdown(self, markdown: str, *, page: Page, config: MkDocsConfig, files: Files):

        # Modify page update date
        # TODO: move date format to config
        # TODO: warn on missing in config
        update_date: datetime = page.meta.get(
            "update", page.meta.get("date", datetime.strptime(page.update_date, "%Y-%m-%d"))
        )
        try:
            page.update_date = update_date.strftime("%Y-%m-%d")
        except AttributeError:
            # TODO: add format fallback
            pass

        if (
            page.file.src_uri in self._hidden_files and not self.config.publish.search_in_hidden
        ) or (page.file.src_uri in self._draft_files and not self.config.publish.search_in_draft):
            page.meta["search"] = {"exclude": True}



# MIT License
#
# Copyright (c) 2023 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import re
from dataclasses import dataclass

# from hashlib import md5
from typing import Dict
from typing import List
from typing import Optional
from typing import cast

from mkdocs.config.defaults import MkDocsConfig
from mkdocs.structure.pages import Page

from mkdocs_publisher._shared import links

# noinspection PyProtectedMember
from mkdocs_publisher._shared.urls import slugify
from mkdocs_publisher.blog.plugin import BlogPlugin

log = logging.getLogger("mkdocs.plugins.publisher.obsidian.backlinks")

HTTP_LINK_RE = re.compile(r"\[([^][\r\n]+)]\((https?://[^][)(\s]+)(#[\w\-.]+)?\)")


@dataclass()
class Link:
    text: str
    destination: str
    title: str
    source: str


class BacklinkLinks:
    def __init__(
        self,
        mkdocs_config: MkDocsConfig,
        backlinks: Dict[str, List[Link]],
    ):
        self._mkdocs_config: MkDocsConfig = mkdocs_config
        self._backlinks: Dict[str, List[Link]] = backlinks

    @staticmethod
    def _build_anchor_link(backlink: str, anchor_link: Optional[str]) -> str:
        """Create backlink anchor link"""

        anchor_link = f"{backlink}{anchor_link}"
        # log.warning(anchor_link)
        for r in (("../", ""), (".md", ""), ("#", ""), ("/", ""), ("_", "-")):
            anchor_link = anchor_link.replace(*r)
        anchor_link = slugify(text=anchor_link)
        # log.error(anchor_link)
        return anchor_link

    @staticmethod
    def _other_link_to_text(match: re.Match) -> str:
        """Return only a text from a link (http or markdown)"""
        return match.group(1)

    def _create_anchor_link(self, match: re.Match):
        """Create a backlink with an additional anchor link"""
        anchor_link = match.group(3) if match.group(3) is not None else ""
        backlink = match.group(2)
        backlink_anchor_link = self._build_anchor_link(backlink=backlink, anchor_link=anchor_link)
        # TODO: create anchor link using slug from meta
        link = f"[{match.group(1)}]({backlink}{anchor_link}){{#{backlink_anchor_link}}}"
        # link_md5 = md5(f"[{match.group(1)}]({backlink}{anchor_link})".encode()).hexdigest()
        # log.error(f"{link_md5} = {link}")
        return link

    def _parse_markdown_link(self, match: re.Match, page: Page, line: str):
        """Parse markdown link"""
        anchor_link = match.group(3) if match.group(3) is not None else ""
        original_link_destination = match.group(2)
        original_link_text = f"[{match.group(1)}]({original_link_destination}{anchor_link})"
        anchor_link = self._build_anchor_link(
            backlink=original_link_destination, anchor_link=anchor_link
        )
        backlink_link = f"{self._mkdocs_config.site_url}{page.url}#{anchor_link}"

        # Convert document link into backlink
        link_replacement = f'<a href="{backlink_link}">{match.group(1)}</a>'
        backlink_text = line.replace(original_link_text, link_replacement)

        # Convert other links into text
        backlink_text = re.sub(links.MD_LINK_RE, self._other_link_to_text, backlink_text)
        backlink_text = re.sub(HTTP_LINK_RE, self._other_link_to_text, backlink_text)

        # Create a backlink with context
        backlink_text = f'<p class="obsidian_backlink">{backlink_text}</p>'

        # Convert relative backlinks into absolute backlinks inside document directory
        original_link_source = page.file.src_uri
        if len(original_link_destination.split("/")) == 1:
            destination_pre = original_link_source.split("/")[:-1]
            destination_pre.append(original_link_destination)
            original_link_destination = "/".join(destination_pre)
        original_link_destination = original_link_destination.replace("../", "")

        link = Link(
            text=backlink_text,
            destination=backlink_link,
            source=original_link_destination,
            title=str(page.title),
        )

        # Get blog temporary files
        temp_blog_files = []
        if "pub-blog" in self._mkdocs_config.plugins:
            blog_plugin: BlogPlugin = cast(BlogPlugin, self._mkdocs_config.plugins["pub-blog"])
            temp_blog_files = blog_plugin.blog_config.temp_files_list

        # Do not add backlink if backlinks points to the same document
        if (
            original_link_source != original_link_destination
            and original_link_source not in temp_blog_files
        ):
            log.debug(
                f"Found backlink to: {match.group(2)}"
                f"{match.group(3) if match.group(3) is not None else ''}"
            )
            if original_link_destination not in self._backlinks:
                self._backlinks[original_link_destination] = [link]
            else:
                link_found = False
                for existing_link in self._backlinks[original_link_destination]:
                    # Add backlink to an existing one
                    if existing_link.title == link.title:
                        link_found = True
                        existing_link.text = f"{existing_link.text}<hr>{link.text}"
                if not link_found:
                    self._backlinks[original_link_destination].append(link)

    def find_markdown_links(self, markdown: str, page: Page):
        # """Find all markdown backlinks"""
        for line in markdown.split("\n"):
            for match in re.finditer(links.MD_LINK_RE, line):
                self._parse_markdown_link(match=match, page=page, line=line)

    def convert_to_anchor_link(self, markdown: str) -> str:
        """Convert backlink to link with an anchor for direct navigation after clicking on it"""
        return re.sub(links.MD_LINK_RE, self._create_anchor_link, markdown)



# MIT License
#
# Copyright (c) 2023 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from typing import cast

from mkdocs.config.defaults import MkDocsConfig

log = logging.getLogger("mkdocs.plugins.publisher._shared.mkdosc_utils")


def get_plugin_config(mkdocs_config: MkDocsConfig, plugin_name: str) -> dict:
    plugins = mkdocs_config.plugins
    if isinstance(plugins, list):
        for plugin in plugins:
            if isinstance(plugin, dict) and list(plugin.keys())[0] == plugin_name:
                return plugin[list(plugin.keys())[0]]
            elif isinstance(plugin, str) and plugin == plugin_name:
                return {}
        raise SystemError("Break")
    else:
        return mkdocs_config.plugins[plugin_name].config


def get_mkdocs_config() -> MkDocsConfig:
    config = MkDocsConfig()
    with open("mkdocs.yml", mode="r") as mkdocs_yml:
        config.load_file(mkdocs_yml)
    return cast(MkDocsConfig, config)
